/*  Richard G Brown */
/*  Plug-in nodes for MQSeries Integrator version 2 */

#include "rgbnode.h"

/* //////////////////////////////////////////////////////////////////// */
/*  helper utilities (cross platform strings, etc) */

#include "rgbnodeutils.h"

/* //////////////////////////////////////////////////////////////////// */
/*  Mandatory Nodal Functions */

/*  CreateNodeContext */

void* CompDecompCreateNodeContext(
				  CciFactory* factoryObject,
				  CciChar*    nodeName,
				  CciNode*    nodeObject)
{
  static char* functionName = (char*)"CompDecompCreateNodeContext()";
  struct CompDecompNodeContext* context;

  int iRc;
  int iRc2;
  int i;

  CciChar* inputDataKey;
  CciChar* outputDataKey;
  CciChar* dataValue;
  CciChar* inValue;
  CciChar* outValue;
  CciChar* specialStringID;
  CciChar* emptyString;

  compLog(1,functionName);

  compLog(0, "Initialising string constants");

  inputDataKey = CciString("inputSource");
  outputDataKey = CciString("outputSource");
  dataValue = CciString("Root.BLOB.BLOB");
  inValue = CciString("in");
  outValue = CciString("out");
  specialStringID=CciString("specialString");
  emptyString=CciString("");

  compLog(0,"Creating context structure");
  context = (struct CompDecompNodeContext*)malloc
    (sizeof(struct CompDecompNodeContext));
  TESTMALLOC(context);
  
  compLog(0,"  context is at %p.", (void*)context);
  compLog(0,"Creating terminals");

  context->outTerminal = cniCreateOutputTerminal
    (&iRc, nodeObject, outValue);
  context->inTerminal = cniCreateInputTerminal
    (&iRc2, nodeObject, inValue);
  
  if (context->outTerminal==CCI_NULL_ADDR ||
      context->inTerminal==CCI_NULL_ADDR ||
      iRc != CCI_SUCCESS || 
      iRc2 != CCI_SUCCESS)
    {
      /*  Failure in terminal creation */
      cciThrowException(&iRc,E_CONFIG,RGB_NODE_TERMINAL_CREATION_FAILED,
			"Failed to create one or more node terminals", 
			(char*)0);
    }

  compLog(0,"  in and out terminals are at %p and %p.",
	  (void*)context->inTerminal,
	  (void*)context->outTerminal);
    
  /*  create a first contextAttribute. Nothing is stored in it */
  /*  yet but it will contain data (and further links) later */

  compLog(0,"Creating first attribute structure");
  context->attribLinkedList=(struct ContextAttribute*)malloc
    (sizeof(struct ContextAttribute));
  TESTMALLOC(context);

  compLog(0,"  first attribute structure is at %p.", 
	  (void*)context->attribLinkedList);
  compLog(0,"Initialising attribute structure");
  initialiseContextAttribute(context->attribLinkedList);

  /*  MQSIv202 configuration manager only sends us non default */
  /*  values. Therefore, it is our responsibility to set all */
  /*  defaults. The attributes in this version are: */

  /*  * input data location */
  /*  * output data location */

  compLog(0,"Putting default values");

  putContextAttribute(inputDataKey, dataValue, context->attribLinkedList, 1);
  putContextAttribute(outputDataKey, dataValue, context->attribLinkedList, 1);
  putContextAttribute(specialStringID, emptyString, context->attribLinkedList, 1);

  compLog(0, "Freeing storage used for character constants");

  free(inputDataKey);
  free(outputDataKey);
  free(dataValue);
  free(inValue);
  free(outValue);
  free(specialStringID);
  free(emptyString);

  compLog(-1,functionName);
  return((void*)(context));
}

/*  DeleteNodeContext */

void CompDecompDeleteNodeContext(CciContext* nodeContext)
{
  static char* functionName = (char*)"CompDecompDeleteNodeContext()";
  int iRc;
  struct CompDecompNodeContext* context;

  compLog(1,functionName);

  compLog(0, "Context is at %p.", (void*)nodeContext);
  context = (struct CompDecompNodeContext*)nodeContext;

  if (context==NULL)
    {
      cciThrowException(&iRc,E_CONFIG,RGB_CONTEXT_ATTRIBUTE_NULL,
			"Function passed null context", (char*)0);
    }

  compLog(0,"Freeing attribute structures rooted at %p.",
	  (void*)context->attribLinkedList);
  freeContextAttribute(context->attribLinkedList);

  compLog(0,"Freeing context structure rooted at %p.",
	  (void*)context);
  free(context);

  compLog(-1,functionName);
  return;
}

/*  GetAttributeName */

int CompDecompGetAttributeName(
			       CciContext* context,
			       int         index,
			       CciChar*    buffer,
			       int         bufsize)
{
  static char* functionName = (char*)"CompDecompGetAttributeName()";

  int error=0; /* Set this to one if request index out of range */

  CciChar* attributeName;
  CciChar* attributeValue;
  struct CompDecompNodeContext* realContext;
  int iRc;

  compLog(1,functionName);
  compLog(0, "Passed a buffer of size %d; searching for index %d.",
	  bufsize, index);
  
  if (context==NULL)
    {
      cciThrowException(&iRc,E_CONFIG,RGB_CONTEXT_ATTRIBUTE_NULL,
			"Function passed null context", (char*)0);
    }

  compLog(0, "  context is at %p.", (void*)context);
  realContext=(struct CompDecompNodeContext*)context;

  compLog(0, "Searching for attribute at index %d.", index);
  attributeValue=getContextAttributeIndex
    (realContext->attribLinkedList, index);
  
  /*  Note that getContextAttributeIndex returns pointer to */
  /*  name so we must not call free */

  if (attributeValue==NULL)
    {  
      /*  assume out of range; */
      /*  let's hope there are no null valid values... */
      compLog(0, "No matching value found");
      error=1;
    }
  else
    {
      /*  copy attributeValue into buffer */
      compLog(0, "Returning value: %s.", 
	      DmbString(attributeValue));
      CciCharNCpy(buffer, attributeValue, bufsize/sizeof(CciChar));
      compLog(0, "Truncated version of string: %s.",
	      DmbString(buffer));
    }

  compLog(-1,functionName);
  return(error);
}

/*  GetAttribute */

int CompDecompGetAttribute(
			   CciContext* context,
			   CciChar*    attrName,
			   CciChar*    buffer,
			   int         bufsize)
{
  static char* functionName = (char*)"CompDecompGetAttribute()";
  int iRc;
  CciChar* value;
  struct CompDecompNodeContext* realContext;
  int rc=0;

  compLog(1,functionName);

  if (context==NULL)
    {
      cciThrowException(&iRc,E_CONFIG,RGB_CONTEXT_ATTRIBUTE_NULL,
			"Function passed null context attribute", (char*)0);
    }

  realContext=(struct CompDecompNodeContext*)context;
  compLog(0,"  context is at %p.", (void*)context);
  compLog(0,"requested key: %s.", DmbString(attrName));

  value=getContextAttribute(attrName, realContext->attribLinkedList);
  if (value!=NULL)
    {
      compLog(0,"Found value: %s.", DmbString(value));
      CciCharNCpy(buffer, value, bufsize/sizeof(CciChar));
      compLog(0,"Truncated version: %s.", DmbString(buffer));
    }
  else
    {
      compLog(0,"no value found");
      rc=1;
    }

  /*  getContextAttribute returns pointer to real value so we must not */
  /*  free value. */

  compLog(-1,functionName);
  return(rc);
}

/*  SetAttribute */

int CompDecompSetAttribute(
			   CciContext* context,
			   CciChar*    attrName,
			   CciChar*    attrValue)
{
  static char* functionName = (char*)"CompDecompSetAttribute()";

  int iRc;
  struct CompDecompNodeContext* realContext;
  int returnValue;

  compLog(1,functionName);

  if (context==NULL)
    {
      cciThrowException(&iRc,E_CONFIG, RGB_CONTEXT_ATTRIBUTE_NULL,
			"Function passed null context", (char*)0);
    }

  realContext=(struct CompDecompNodeContext*)context;
  compLog(0,"  context is at %p.", (void*)context);

  compLog(0,"Putting key: %s; value: %s.", DmbString(attrName),
	  DmbString(attrValue));
  
  /* 0 flag means we don't create attributes that didn't
     already exist */

  returnValue = putContextAttribute(attrName, attrValue, 
		realContext->attribLinkedList,0);

  compLog(-1,functionName);

  return(returnValue);
}

/*  CompressorEvaluate */

void CompressorEvaluate(
			CciContext* context,
			CciMessage* pDestinationList,
			CciMessage* pExceptionList,
			CciMessage* pMessage)
{
  static char* functionName = (char*)"CompressorEvaluate()";
  struct messageElements* msgElts;

  CciValueType elementValueType;

  int iRc;
  int rc;

  int i; /* it would be preferable if these variables were */
  int j; /* unsigned but the MQSI header files specify signed int for */
         /* important data structures */

  /*  we create the two byte array structs on the stack. */
  /*  The byte arrays they contain are created on the heap. */
  /*  Be careful with deletion... */

  struct CciByteArray inputData;
  struct CciByteArray outputData;

  struct CompDecompNodeContext* nodeContext;
  CciTerminal* propagationTerminal;
  
  CciChar* specialStringID; /* This is the string passed across the 
			     data transformation interface to allow
			     an algorithm-specific string to be passed to
			     the relevant algorithm without recoding this file */ 
  CciChar* specialString;

  compLog(1,functionName);
  
  if (context==NULL)
    {
      cciThrowException(&iRc,E_RECOV,RGB_CONTEXT_ATTRIBUTE_NULL,
			"Function passed null context", (char*)0);
    }

  compLog(0, "  context is at %p.", (void*)context);
  nodeContext = (struct CompDecompNodeContext*)context;
  
  compLog(0, "Extracting free-form special string");
  specialStringID=CciString("specialString");
  specialString=getContextAttribute(specialStringID, nodeContext->attribLinkedList);

  compLog(0, "Allocating storage for message elements");
  msgElts = (struct messageElements*)
    malloc(sizeof(struct messageElements));
  TESTMALLOC(msgElts);

  compLog(0, "Preparing message elements structure");
  prepareMessage(pMessage,
		 nodeContext,
		 msgElts);

  propagationTerminal = nodeContext->outTerminal;

  inputData.pointer=NULL;
  compLog(0, "Determining size of input data.");
  
  compLog(0, "First determining what kind of element this is...");	
  elementValueType = cniElementValueType(&iRc, msgElts->inputElement);
  compLog(0, "Element type was %d.", elementValueType);
  
  compLog(0, "Allocating storage for data - we will soon free this");	
  inputData.pointer=malloc(10);
  TESTMALLOC(inputData.pointer);
  inputData.size=10;
  inputData.size = cniElementByteArrayValue(&iRc,
				       msgElts->inputElement,
				       &inputData);
  
  compLog(0, "Return code was %d; inputData.size is %d", iRc, inputData.size);
  
  if (iRc != CCI_SUCCESS && iRc != CCI_BUFFER_TOO_SMALL)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_NOT_A_BYTE_ARRAY,
			"Failed to determine size of input byte array; "
			"Check that the input location contains a byte "
			"array.", (char*)0);
    }

  compLog(0, "Input data size: %d", inputData.size);
  compLog(0, "Creating space for inputData buffer to prevent "
	          "cniElementByteArrayValue messing up");
  /* I guess realloc would be the way to go here... */
  free(inputData.pointer);
  inputData.pointer = malloc(inputData.size);
  TESTMALLOC(inputData.pointer);

  compLog(0, "Copying input data into buffer");
  /* ignore return code */ cniElementByteArrayValue(&iRc,
					       msgElts->inputElement,
					       &inputData);
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_FAILED_GET_BYTE_ARRAY,
			"Failed to copy input byte array into buffer",
			(char*)0);
    }					       

  /*  We now create the out buffer. We know it can be larger than the input */
  /*  data so we allocate enough space for the output to be as large */
  /*  as but no larger than the input data. If the compressor complains that */
  /*  there is not enough output space, this means that the compression */
  /*  has failed to shrink the data (perfectly possible). It will allocate more */
  /*  space and try again. */

  outputData.size = inputData.size;

  compLog(0,"Creating buffer of size %d for outputData", 
	  outputData.size);
  outputData.pointer = malloc(outputData.size);
  TESTMALLOC(outputData.pointer);
  
  /*  If the input data is of zero length, we skip the compression phase */
  if (inputData.size>0)
    {
      /***********************************
        This is where we do the real work
       ***********************************/
      compressData(&inputData, &outputData, specialString);	
    }
  else
    {
      compLog(0, "input data had zero length; setting output size to 0");
      outputData.size=0;
    }

  compLog(0, "resizing output buffer to new size: %d",
	  outputData.size);
  
  outputData.pointer=realloc(outputData.pointer, outputData.size);
  TESTMALLOC(outputData.pointer);

  compLog(0, "Placing output buffer into message tree");
  cniSetElementByteArrayValue(&iRc,
			      msgElts->outputElement,
			      &outputData);
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}

      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"CompressorEvaluate failed to set outbound "
			"message value", "Failed to set output data "
			"into output message.", (char*)0);
    }
  
  compLog(0, "Propagating message to output terminal");
  rc = cniPropagate(&iRc,
		    propagationTerminal,
		    pDestinationList,
		    pExceptionList,
		    msgElts->outputMessage);

  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Error propagating message", 
			"Error propagating message",
			(char*)0);
    }
  
  compLog(0, "Tidying up. Freeing input and output data pointers");
  free(inputData.pointer);
  free(outputData.pointer);

  /*  no need to delete the CciByteArray structures since they were */
  /*  created on the stack */

  compLog(0, "Deleting message structure");
  cniDeleteMessage(&iRc,
		   msgElts->outputMessage);

  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}

      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Failed to delete output message",
			"Failed to delete output message",
			(char*)0);
    }

  compLog(0, "Freeing message elements structure");
  free(msgElts);
  free(specialStringID);
  
  compLog(-1,functionName);
  return;
}

/*  DecompressorEvaluate */

void DecompressorEvaluate(
			  CciContext* context,
			  CciMessage* pDestinationList,
			  CciMessage* pExceptionList,
			  CciMessage* pMessage)
{
  static char* functionName = (char*)"DeompressorEvaluate()";

  struct messageElements* msgElts;

  int iRc;
  int rc;
  int i;

  /*  we create the two byte array structs on the stack. */
  /*  The byte arrays they contain will be created on the heap */

  struct CciByteArray inputData;
  struct CciByteArray outputData;

  struct CompDecompNodeContext* nodeContext;
  CciTerminal* propagationTerminal;

  CciChar* specialStringID; /* This is the string passed across the 
			     data transformation interface to allow
			     an algorithm-specific string to be passed to
			     the relevant algorithm without recoding this file */ 
  CciChar* specialString;

  compLog(1,functionName);

  if (context==NULL)
    {
      cciThrowException(&iRc,E_RECOV,RGB_CONTEXT_ATTRIBUTE_NULL,
			"Null context provided", (char*)0);
    }

  compLog(0, "Context is at %p.", (void*)context);  
  nodeContext = (struct CompDecompNodeContext*)context;
  
  compLog(0, "Extracting free-form special string");
  specialStringID=CciString("specialString");
  specialString=getContextAttribute(specialStringID, nodeContext->attribLinkedList);
  
  compLog(0, "Allocating storage for message elements");
  msgElts = (struct messageElements*)
    malloc(sizeof(struct messageElements));
  TESTMALLOC(msgElts);

  compLog(0, "Preparing message elements structure");
  prepareMessage(pMessage,
		 nodeContext,
		 msgElts);

  /*  We expect structure on the output byte stream: */
  /*  first four bytes contain the length of the input data */
  /*  in BIG-ENDIAN FORM - Most Significant Byte First */
  /*  The compressed data comes after this. */

  propagationTerminal = nodeContext->outTerminal;
  
  compLog(0, "Determining size of input byte array. Temporarily allocating 10 "
	          "bytes so cniElementByteArrayValue doesn't mess up");
  /* I guess realloc would be the way to go here... */
  inputData.size=10;
  inputData.pointer=malloc(10);
  TESTMALLOC(inputData.pointer);
  inputData.size = cniElementByteArrayValue(&iRc,
				       msgElts->inputElement,
				       &inputData);
  
  if (iRc != CCI_SUCCESS && iRc != CCI_BUFFER_TOO_SMALL)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_NOT_A_BYTE_ARRAY,
			"Failed to determine size of input element. "
			"This is probably caused by target syntax "
			"element not being a byte-array", (char*)0);
    }

  compLog(0, "input data size was %d; allocating memory", 
	  inputData.size);
  free(inputData.pointer);
  inputData.pointer = malloc(inputData.size);
  TESTMALLOC(inputData.pointer);

  compLog(0, "Copying data into buffer");
  /* ignore return code */ cniElementByteArrayValue(&iRc,
					       msgElts->inputElement,
					       &inputData);
  
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      
      cciThrowException(&iRc,E_RECOV,RGB_NOT_A_BYTE_ARRAY,
			"Failed to copy data into a buffer. "
			"This is probably caused by target syntax "
			"element not being a byte-array", (char*)0);    }					       

  /*  We now create the output buffer. We use the first 4 bytes */
  /*  of the incoming message to tell us how large the uncompressed */
  /*  message is.  */

  outputData.size = inputData.size;
  compLog(0, "allocating memory to hold output message");
  outputData.pointer = malloc(inputData.size);
  TESTMALLOC(outputData.pointer);

  if (inputData.size>0)
    {
      /***********************************
        This is where we do the real work
       ***********************************/
      decompressData(&inputData, &outputData, (CciChar*)0);
    }
  else
    {
      compLog(0, "Output message size calculated as zero so doing nothing");
    }

  outputData.pointer=realloc(outputData.pointer, outputData.size);
  TESTMALLOC(outputData.pointer);

  compLog(0, "Placing output byte array in message tree");
  cniSetElementByteArrayValue(&iRc,
			      msgElts->outputElement,
			      &outputData);
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Failed to set output data into message tree",
			"Failed to set output data into message tree",
			(char*)0);
    }
  
  compLog(0, "propagating to output terminal");
  rc = cniPropagate(&iRc,
		    propagationTerminal,
		    pDestinationList,
		    pExceptionList,
		    msgElts->outputMessage);

  if (rc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Failed to propagate message to output terminal",
			"Failed to propagate message to output terminal",
			(char*)0);
    }

  compLog(0, "Freeing input and output data memory");
  free(inputData.pointer);
  free(outputData.pointer);

  /*  no need to delete the CciByteArray structures since they were */
  /*  created on the stack */
  
  compLog(0, "Freeing message structure");
  cniDeleteMessage(&iRc,
		   msgElts->outputMessage);

  if (iRc != CCI_SUCCESS)
    {
      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Failed to delete output message object",
			"Failed to delete output message object",
			(char*)0);
    }

  compLog(0, "Freeing message elements structure");
  free(msgElts);
  free(specialStringID);

  compLog(-1,functionName);
  return;
}

/* //////////////////////////////////////////////////////////////////// */
/*  Mandatory Factory Function */

/*  bipGetMessageflowNodeFactory */

void LilFactoryExportPrefix * LilFactoryExportSuffix 
bipGetMessageflowNodeFactory()
{
  static char* functionName = (char*)"bipGetMessageflowNodeFactory()";
		/* Remove the comments on the next two lines to compile
		   against Websphere MQ Integrator 2.1 */
  static CNI_VFT compressorVirtualFunctionTable /* = { CNI_VFT_DEFAULT } */ ;
  static CNI_VFT decompressorVirtualFunctionTable /* = { CNI_VFT_DEFAULT } */ ;

  int iRc;
  int iRc2; /*  to save effort checking errors twice */

  CciFactory* factoryObject;

  static CciChar* factoryName;
  static CciChar* compressorNodeClass;
  static CciChar* decompressorNodeClass;

  initMessageSourcePointer(); /* This should really be done as part 
			         the initialisaton of the rgbnodeutils shared object */
   
  compLog(1,functionName);

  compLog(0, "Initialising string constants");

  factoryName = CciString("CompressionFactory");
  compressorNodeClass = CciString("CompressionNode");
  decompressorNodeClass = CciString("DecompressionNode");  
  
  /* Comment out the following two memset calls if building for
     Websphere MQ Integrator 2.1 */

  memset(&compressorVirtualFunctionTable, 0, 
	 sizeof(CNI_VFT));
  memset(&decompressorVirtualFunctionTable, 0, 
	 sizeof(CNI_VFT));
  
  compressorVirtualFunctionTable.iFpCreateNodeContext
    = CompDecompCreateNodeContext;
  compressorVirtualFunctionTable.iFpDeleteNodeContext
    = CompDecompDeleteNodeContext;
  compressorVirtualFunctionTable.iFpGetAttributeName
    = CompDecompGetAttributeName;
  compressorVirtualFunctionTable.iFpSetAttribute
    = CompDecompSetAttribute;
  compressorVirtualFunctionTable.iFpGetAttribute
    = CompDecompGetAttribute;
  compressorVirtualFunctionTable.iFpEvaluate
    = CompressorEvaluate;

  decompressorVirtualFunctionTable.iFpCreateNodeContext
    = CompDecompCreateNodeContext;
  decompressorVirtualFunctionTable.iFpDeleteNodeContext
    = CompDecompDeleteNodeContext;
  decompressorVirtualFunctionTable.iFpGetAttributeName
    = CompDecompGetAttributeName;
  decompressorVirtualFunctionTable.iFpSetAttribute
    = CompDecompSetAttribute;
  decompressorVirtualFunctionTable.iFpGetAttribute
    = CompDecompGetAttribute;
  decompressorVirtualFunctionTable.iFpEvaluate
    = DecompressorEvaluate;

  compLog(0, "Creating node factory with name %s.", DmbString(factoryName));
  factoryObject = cniCreateNodeFactory(&iRc, factoryName);

  if (factoryObject == CCI_NULL_ADDR)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}

      cciThrowException(&iRc,E_CONFIG,RGB_NULL_POINTER_EXCEPTION,
			"Null pointer exception", (char*)0);
    }
  
  compLog(0, "Defining compressor node class: %s", 
	  DmbString(compressorNodeClass));
  cniDefineNodeClass(&iRc, factoryObject, compressorNodeClass, 
		     &compressorVirtualFunctionTable);
  compLog(0, "Return code was %d", iRc);
  
  compLog(0, "Defining decompressor node class: %s", 
	  DmbString(decompressorNodeClass));
  cniDefineNodeClass(&iRc2, factoryObject, decompressorNodeClass, 
		     &decompressorVirtualFunctionTable);
  compLog(0, "Return code was %d", iRc2); 
  
  if ((iRc != CCI_SUCCESS) || (iRc2 != CCI_SUCCESS))
    {      
      if (iRc == CCI_EXCEPTION || iRc2 == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_CONFIG,RGB_NODE_CLASS_CREATION_FAILED,
			"Failed to create one of both of compressor "
			"and decompressor nodes", (char*)0);
    }

  compLog(0, "Freeing character constants");

  free(factoryName);
  free(compressorNodeClass);
  free(decompressorNodeClass);

  compLog(-1,functionName);
  return((void*)factoryObject);
}

